con
  cntMin = 400                  'Minimum waitcnt value to prevent lock-up
  Max_uS = 2350                 'longest valid servo pulse (in microseconds)
  Min_uS = 300                  'shortest valid servo pulse (in microseconds)
  Precision = 5                 'how small of a step the search uses
  Start_offset = 100            'how far to offset form the assumed position to start
  PP_wait_div = 200             '1 / PP_wait_div seconds between search pulses
  Num_ch = 4                    'number of servos (physical_channels - 1)
  Start_byte = 33               'serial communication start byte
  
var                             
  'these variables are unique to each instance of this object.  
  long error_code               'stores the most recent status code of the servo
  long cog                      'stores the cog number used, and weather it is still active

obj
  serial : "fullduplexserial.spin"
  {eventually this opject will need to be modified so that all the variables are in the DAT section of the code.  This mod will
  allow one cog running fullDuplexSerial to service any number of cogs running Servo_diagnostics, probably should add lock setting
  and checking code to it too.  It'd be best if only one person used the serial port at a time.  
  this should only be started in the Init method, this method is run ONCE for all instances of the Servo_diagnostics object}                     

pub Init | temp, index, pin 
{{initalise all the timing for this monster code.  
Init needs to be called at startup and anytime the clock speed changes
Only needs to be called once for all instances of the object.}}
   us := clkfreq / 1_000_000                  ' Clock cycles for 1 us

   'setup the serial port
   'once full_dup ismodified this should only happen once for ALL cogs using it
   'I probably should have the port pins, etc passed in as paramiters
   serial.start(27, 26, 0, 9600)
            

Pub Pulseout_uS(Pin, Duration) | SyncCNT
''generate a pulse on Pin, Duration microseconds long.
  Duration := Min_uS #> Duration <# Max_uS     'limit Duration to valid values                                                     
  dira[pin]~~                                  ' Set to output                                         
  SyncCNT := cnt + 1000                        'snag value of cnt + some time
  waitcnt(SyncCNT)                             'sync to cnt
  !outa[pin]                                   'set Pin to opposite state                                 
  {do math while we wait for CNT to reach our target}                
  SyncCNT += ((Duration * us) #> cntMin)       'modify sync point to a future time. duration*(clk cycles for us)
  waitcnt(SyncCNT)                             'wait until clk gets there                             
  !outa[pin]                                   'return Pin to orig. state

con
  Offset_up_err = 50            'offset error of up search method

pub Find_pos_up(Pin,AssumedPos):RealPos | error
{{search for the current position of the servo.
assumes signal pin is Pin and mode pin is Pin+1}}
  dira[Pin+1]~                  'make the mode pin an input
  error := 0                    'counter for how many time's search has happened
  RealPos := AssumedPos - Start_offset - Offset_up_err    'start the search just below AssumedPos
  {setup for main loop}
  Pulseout_uS(Pin,RealPos)      'pulse the servo
  waitcnt(clkfreq/PP_wait_div + cnt)    'wait for 5ms
  if ina[Pin+1]                 'if servo's internal position matches the command
    RealPos += Offset_up_err    'adjust for offset error
    return                      'return RealPos
  {main loop}
  repeat                        'loop through the avalable positions
    RealPos += Precision                'increment tested position by 5us
    if RealPos => Max_uS        'check if the command pulse is too long
      RealPos := Min_uS         'start over at the min pulse length
      error++
    Pulseout_uS(Pin,RealPos)    'pulse the servo
    waitcnt(clkfreq/PP_wait_div + cnt)  'wait for 5ms
  until ina[Pin+1] or (error > 3)    'I'm stuck in here till I match positions with the servo's internals or I give up
  if error > 3                  'did I give up?
    RealPos := -1               'something's not right.  return an error code.
  RealPos += Offset_up_err      'adjust position for offset error
  dira[Pin+1]~~                 'we're done, make the mode pin an output

con
  Offset_down_err = 25            'offset error of up search method

pub Find_pos_down(Pin,AssumedPos):RealPos | error
{{search for the current position of the servo.
assumes signal pin is Pin and mode pin is Pin+1
for some reson this is more glitchy than the up search}}
  dira[Pin+1]~                  'make the mode pin an input
  error := 0                    'counter for how many time's search has happened
  RealPos := AssumedPos + Start_offset - Offset_down_err    'start the search just abovve AssumedPos
  {setup for main loop}
  Pulseout_uS(Pin,RealPos)      'pulse the servo
  waitcnt(clkfreq/PP_wait_div + cnt)    'wait for 5ms
  if ina[Pin+1]                 'if servo's internal position matches the command
    RealPos += Offset_down_err    'adjust for offset error
    return                      'return RealPos
  {main loop}
  repeat                        'loop through the avalable positions
    RealPos -= Precision                'decrement tested position by 5us
    if RealPos =< Min_uS        'check if the command pulse is too short
      RealPos := Max_uS         'start over at the max pulse length
      error++
    Pulseout_uS(Pin,RealPos)    'pulse the servo
    waitcnt(clkfreq/PP_wait_div + cnt)  'wait for 5ms
  until ina[Pin+1] or (error > 3)    'I'm stuck in here till I match positions with the servo's internals or I give up
  if error > 3                  'did I give up?
    RealPos := -1               'something's not right.  return an error code.
  RealPos += Offset_down_err      'adjust position for offset error
  dira[Pin+1]~~                 'we're done, make the mode pin an output

Pub Move_servo(Target_pos, Channel):temp2 | signal_pin, Lo_limit, Hi_limit, speed, temp, real_pos, old_real_pos, old_error_code
{{decode Target_pos (0 to 255) into the appropriate pulse for Channel.
Move the servo to the above position at Speed_of_movement
Verify that the servo reached Target_pos
Set an object golbal variable with the appropriate error_code
Also returns the error_code if the function is run inline}}

  'clear error code to indicate that i'm active
  error_code := 0

  'extract important data from the central storage area
  temp := servo_dat_point(Channel)                      'read in the pointer to my data
  repeat while long[temp] < 0                           'wait for the servo to be free
  signal_pin := long[temp]                              'extract the pin # I'm using
  !long[temp]                                           'mark that i'm using the servo
  speed := long[temp][1]                                'extract how fast I should move (uS/loop)
  Lo_limit := long[temp][2]                             'extract one of my motion limits (uS)
  Hi_limit := long[temp][3]                             'extract the other motion limit (uS)
  old_real_pos := long[temp][4]                         'extract my prevously found position (uS)
  old_error_code := long[temp][5]                       'extract my last error code (not using it at the moment)
  
  'catch a startup glitch
  if old_real_pos == 0                                  'if this is the first run
    old_real_pos := find_pos_up(signal_pin,1200)        'find my current position and assign it to old_real_pos

  'scale Target_pos to this axis
  Target_pos := 0 #> Target_pos <# 255                  'limit range
  Target_pos := (Hi_limit - Lo_limit) * Target_pos / 255 + Lo_limit             're-scale

  'activate the mode pin
  dira[signal_pin+1]~~   'enable servo drive with my special circuit
  outa[signal_pin+1]~~   'enable servo drive with my special circuit
  
  'move to Target_pos in steps of Speed
  repeat temp from old_real_pos to Target_pos step speed
    Pulseout_us(signal_pin,temp)
    waitcnt(clkfreq/60 + cnt)
  repeat 4                      'make shure i've reached the target position
    pulseout_us(signal_pin,Target_pos)
    waitcnt(clkfreq/60 + cnt)  

  'deactivate the mode pin
  dira[signal_pin+1]~   'disable servo drive with my special circuit
  outa[signal_pin+1]~   'disable servo drive with my special circuit
  
  'Search for current position and update error_code
  real_pos := Find_pos_up(signal_pin,Target_pos)        'find current position
  temp := ||(Target_pos - real_pos)                     'how far off am I?
  if temp =< 10                 'everything's fine
    error_code += 1
  if temp > 10                  'short of my goal
    error_code += 8
  if real_pos < 0               'search for my current position failed
    error_code += 2
  temp := ||(old_real_pos - real_pos)                   'some math
  if temp =< 10                 'I didn't even move!
    error_code += 4

  'update history variables in the servo infor array
  temp := servo_dat_point(Channel)  'pre calculate the pointer and store it in Temp                                            
  long[temp][4] := real_pos         'update the old position
  long[temp][5] := error_code       'update the old error code
  !long[temp]                       'mark that the servo is free

  'return the latest error code
  temp2 := error_code

pub get_error:temp              'return the current error code for this object
  temp := error_code

Pub servo_dat_point(Channel):pointer                    'return the pointer into the servo array for Channel
  Channel := 0 #> Channel <# Num_ch                     'limit Channel to a valid servo channel 
  pointer := (Channel * 6 * 4) + @servo_dat             'add in the appropriate offset
  
pub commander | temp, data[3], i, status
{{this is the main command processor.  it watches the serial port for a start byte.
Grabs a packet.  Verifies the packet.  then interperts the command and data.  afterwards it returns
a packet containing the finished command and the status_code of that completed command}}
{assumes Init has been run once and FullDuplexSerial is running}
{since I just need a "move" command commands are a binary value from 0 to Num_ch followed by the position byte 0 to 255}
{other debug, setup, and training commands can be added above these}
{configuring different start bytes allows addressing multiple controllers on one serial port}
{one start byte should be reserved as the "reset" byte.  just send more than three of this byte to
reset the state of all controllers}

  repeat
    temp := serial.rx           'grab a byte
    if temp == Start_byte       'is it the start byte?
      repeat i from 0 to 2      'get the rest of the packet
        temp += data[i] := serial.rx
      temp -= data[2]          'remove x-mitted checksum from accumulator
      temp //= 256             'compute a new checksum
      if data[2] == temp                                'does it match the sent sum?
        status := 0             'clear the status variable
        case data[0]
          0..Num_ch:            'move commands
            status := move_servo(data[1],data[0])        'move the servo and update the status
          other:                'don't know what u sent crazy man!
            status += 16                                'bad command
        serial.tx(Start_byte)   'start byte
        serial.tx(data[0])      'echo command
        serial.tx(status)       'send status code
        temp := (Start_byte + data[0] + status) // 256
        serial.tx(temp)         'send a checksum


Dat
{{ single structure containing all the setup and history information
for each servo axis.  shared by ALL of the instances of this object}}

us            long      0-0                             'store timing variable in common for ALL instances of this object
servo_dat     long      0, 5,   925, 1655, 0-0, 0-0    'current data for 5 axis of servos.  (no real limit actually)
              long      2, 7,  1000, 1500, 0-0, 0-0
              long      4, 10, 1000, 1500, 0-0, 0-0
              long      6, 13, 1000, 1500, 0-0, 0-0
              long      8, 15, 1000, 1500, 0-0, 0-0